@@ -1,6 +1,5 @@
 
 import pytest
-import pytz
 
 import pkg_resources
 
@@ -8,18 +7,12 @@ from datetime import date, datetime, timedelta, time
 import icalendar
 
 from khal.khalendar import backend
-from khal.khalendar.event import LocalizedEvent
+from khal.khalendar.event import LocalizedEvent, EventStandIn
 from khal.khalendar.exceptions import OutdatedDbVersionError, UpdateFailed
 
-from .aux import _get_text
-
-BERLIN = pytz.timezone('Europe/Berlin')
-LONDON = pytz.timezone('Europe/London')
-SAMOA = pytz.timezone('Pacific/Samoa')
-SYDNEY = pytz.timezone('Australia/Sydney')
-LOCALE_BERLIN = {'local_timezone': BERLIN, 'default_timezone': BERLIN}
-LOCALE_SAMOA = {'local_timezone': SAMOA, 'default_timezone': SAMOA}
-LOCALE_SYDNEY = {'local_timezone': SYDNEY, 'default_timezone': SYDNEY}
+from .utils import _get_text, \
+    BERLIN, LONDON, SYDNEY, \
+    LOCALE_BERLIN, LOCALE_SYDNEY
 
 calname = 'home'
 
@@ -51,6 +44,16 @@ def test_event_rrule_recurrence_id():
     assert events[4].start == BERLIN.localize(datetime(2014, 7, 28, 7, 0))
     assert events[5].start == BERLIN.localize(datetime(2014, 8, 4, 7, 0))
 
+    events = dbi.get_localized(
+        BERLIN.localize(datetime(2014, 6, 30, 0, 0)),
+        BERLIN.localize(datetime(2014, 8, 26, 0, 0)),
+        minimal=True,
+    )
+    events = list(events)
+    assert len(events) == 6
+    for event in events:
+        assert isinstance(event, EventStandIn)
+
 
 def test_event_different_timezones():
     dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
@@ -63,6 +66,8 @@ def test_event_different_timezones():
     event = events[0]
     assert event.start_local == LONDON.localize(datetime(2014, 4, 9, 14))
     assert event.end_local == LONDON.localize(datetime(2014, 4, 9, 19))
+    assert event.start == LONDON.localize(datetime(2014, 4, 9, 14))
+    assert event.end == LONDON.localize(datetime(2014, 4, 9, 19))
 
     # no event scheduled on the next day
     events = dbi.get_localized(BERLIN.localize(datetime(2014, 4, 10, 0, 0)),
@@ -79,6 +84,8 @@ def test_event_different_timezones():
     event = events[0]
     assert event.start_local == SYDNEY.localize(datetime(2014, 4, 9, 23))
     assert event.end_local == SYDNEY.localize(datetime(2014, 4, 10, 4))
+    assert event.start == LONDON.localize(datetime(2014, 4, 9, 14))
+    assert event.end == LONDON.localize(datetime(2014, 4, 9, 19))
 
     # the event spans midnight Sydney, therefor it should also show up on the
     # next day
@@ -171,6 +178,50 @@ def test_event_rrule_recurrence_id_update_with_exclude():
     assert events[4].start == BERLIN.localize(datetime(2014, 8, 4, 7, 0))
 
 
+def test_event_recuid_no_master():
+    """
+    test for events which have a RECUID component, but the master event is
+    not present in the same file
+    """
+    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
+    dbi.update(_get_text('event_dt_recuid_no_master'),
+               href='12345.ics', etag='abcd', calendar=calname)
+    events = dbi.get_floating(
+        datetime(2017, 3, 1, 0, 0), datetime(2017, 4, 1, 0, 0),
+    )
+    events = sorted(events, key=lambda x: x.start)
+    assert len(events) == 1
+    assert events[0].start == datetime(2017, 3, 29, 16)
+    assert events[0].end == datetime(2017, 3, 29, 16, 25)
+    assert events[0].format(
+        '{title}', relative_to=date(2017, 3, 29)
+    ) == 'Infrastructure Planning\x1b[0m'
+
+
+def test_event_recuid_rrule_no_master():
+    """
+    test for events which have a RECUID and a RRULE component, but the master event is
+    not present in the same file
+    """
+    dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
+    dbi.update(
+        _get_text('event_dt_multi_recuid_no_master'),
+        href='12345.ics', etag='abcd', calendar=calname,
+    )
+    events = dbi.get_floating(
+        datetime(2010, 1, 1, 0, 0), datetime(2020, 1, 1, 0, 0),
+    )
+    events = sorted(events, key=lambda x: x.start)
+    assert len(list(events)) == 2
+    assert events[0].start == datetime(2014, 6, 30, 7, 30)
+    assert events[0].end == datetime(2014, 6, 30, 12, 0)
+    assert events[1].start == datetime(2014, 7, 7, 8, 30)
+    assert events[1].end == datetime(2014, 7, 7, 12, 0)
+    events = dbi.search('VEVENT')
+    events = sorted(events, key=lambda x: x.start)
+    assert len(list(events)) == 2
+
+
 def test_no_valid_timezone():
     dbi = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
     dbi.update(_get_text('event_dt_local_missing_tz'),
@@ -181,6 +232,9 @@ def test_no_valid_timezone():
     assert len(events) == 1
     event = events[0]
     assert event.start == BERLIN.localize(datetime(2014, 4, 9, 9, 30))
+    assert event.end == BERLIN.localize(datetime(2014, 4, 9, 10, 30))
+    assert event.start_local == BERLIN.localize(datetime(2014, 4, 9, 9, 30))
+    assert event.end_local == BERLIN.localize(datetime(2014, 4, 9, 10, 30))
 
 
 def test_event_delete():
@@ -226,6 +280,7 @@ def test_this_and_prior():
     with pytest.raises(UpdateFailed):
         dbi.update(event_rrule_this_and_prior, href='12345.ics', etag='abcd', calendar=calname)
 
+
 event_rrule_this_and_future_temp = """
 BEGIN:VCALENDAR
 BEGIN:VEVENT
@@ -274,7 +329,7 @@ def test_event_rrule_this_and_future():
 
     assert str(events[0].summary) == 'Arbeit'
     for num, event in enumerate(events[1:]):
-        event.raw
+        assert event.raw  # just making sure we don't raise any exception
         assert str(event.summary) == 'Arbeit (lang)'
 
 
@@ -310,6 +365,7 @@ def test_event_rrule_this_and_future_multi_day_shift():
     for event in events[1:]:
         assert str(event.summary) == 'Arbeit (lang)'
 
+
 event_rrule_this_and_future_allday_temp = """
 BEGIN:VCALENDAR
 BEGIN:VEVENT
@@ -359,6 +415,15 @@ def test_event_rrule_this_and_future_allday():
     for event in events[1:]:
         assert str(event.summary) == 'Arbeit (lang)'
 
+    events = list(dbi.get_floating(
+        datetime(2014, 4, 30, 0, 0),
+        datetime(2014, 9, 27, 0, 0),
+        minimal=True,
+    ))
+    assert len(events) == 6
+    for event in events:
+        assert isinstance(event, EventStandIn)
+
 
 def test_event_rrule_this_and_future_allday_prior():
     event_rrule_this_and_future_allday_prior = \
@@ -475,6 +540,7 @@ def test_calc_shift_deltas():
     assert (timedelta(hours=2), timedelta(hours=4, minutes=30)) == \
         backend.calc_shift_deltas(recuid_this_future_duration)
 
+
 event_a = """BEGIN:VEVENT
 UID:123
 SUMMARY:event a
@@ -548,7 +614,7 @@ def test_update_one_should_not_affect_others():
     assert len(events) == 1
 
 
-def test_zuluv_events():
+def test_zulu_events():
     """test if events in Zulu time are correctly recognized as locaized events"""
     db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
     db.update(_get_text('event_dt_simple_zulu'), href='event_zulu', calendar=calname)
@@ -619,6 +685,13 @@ BDAY:19710311
 END:VCARD
 """
 
+card_29thfeb = """BEGIN:VCARD
+VERSION:3.0
+FN:leapyear
+BDAY:20000229
+END:VCARD
+"""
+
 card_no_year = """BEGIN:VCARD
 VERSION:3.0
 FN:Unix
@@ -665,6 +738,31 @@ def test_birthdays():
     assert events[0].summary == 'Unix\'s 45th birthday'
 
 
+def test_birthdays_update():
+    """test if we can update a birthday"""
+    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
+    db.update_birthday(card, 'unix.vcf', calendar=calname)
+    db.update_birthday(card, 'unix.vcf', calendar=calname)
+
+
+def test_birthdays_29feb():
+    """test how we deal with birthdays on 29th of feb in leap years"""
+    db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
+    db.update_birthday(card_29thfeb, 'leap.vcf', calendar=calname)
+    events = list(
+        db.get_floating(datetime(2004, 1, 1, 0, 0), datetime(2004, 12, 31))
+    )
+    assert len(events) == 1
+    assert events[0].summary == 'leapyear\'s 4th birthday (29th of Feb.)'
+    assert events[0].start == date(2004, 2, 29)
+    events = list(
+        db.get_floating(datetime(2005, 1, 1, 0, 0), datetime(2005, 12, 31))
+    )
+    assert len(events) == 1
+    assert events[0].summary == 'leapyear\'s 5th birthday (29th of Feb.)'
+    assert events[0].start == date(2005, 3, 1)
+
+
 def test_birthdays_no_year():
     db = backend.SQLiteDb([calname], ':memory:', locale=LOCALE_BERLIN)
     assert list(db.get_floating(start, end)) == list()
